// @ myth - server - invent.C
//
// Myth is a multiuser screen oriented adventure game.
// Copyright (C) 1989 Per Abrahamsen
// 
// This file is part of Myth.
// 
// Myth is free software; you can redistribute it and/or modify it under
// the terms of the GNU General Public License as published by the Free
// Software Foundation; either version 1, or (at your option) any later
// version.
// 
// Myth is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
// FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
// for more details.
// 
// You should have received a copy of the GNU General Public License
// along with Myth; see the file COPYING.  If not, write to the Free
// Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
//
// My email address is: amanda@iesd.auc.dk

// @@ include

#include "server/invent.h"
#include "server/player.h"
#include "common/error.h"
#include "common/page.h"
#include "server/rnd.h"
#include "server/specify.h"
#include "common/format.h"

// @@ INVENTORY

// @@@ void INVENTORY::CleanUp()
//
// Sort and reduce the list, by adding equal objects together.
// Use with care!

void INVENTORY::CleanUp()
{
    list.sort();

    OBJECT* previous = NULL;

    reset();
    while(remain() > 0)
    {
	if(previous != NULL &&
	   previous->compare(current()) == EQUAL &&
	   MayAdd(current()) && MayAdd(previous))
	{
	    previous->add(current());
	    list.erase();
	}
	else
	{
	    previous = current();
	    next();
	}
    }
}

// @@@ void INVENTORY::SetLetter(symbol_type symbol, char letter)
//
// Set letter in bitmap. ASCII dependend

void INVENTORY::SetLetter(symbol_type symbol, char letter)
{
    if(isupper(letter))
	upper[SymbolToInt(symbol)] |= (1 << letter - 'A');
    else if(islower(letter))
	lower[SymbolToInt(symbol)] |= (1 << letter - 'a');
}

// @@@ bool INVENTORY::CheckLetter(symbol_type symbol, char letter)

bool INVENTORY::CheckLetter(symbol_type symbol, char letter)
{
    if(isupper(letter))
	return (upper[SymbolToInt(symbol)] & (1 << letter - 'A')) != 0;
    else if(islower(letter))
	return (lower[SymbolToInt(symbol)] & (1 << letter - 'a')) != 0;
    else
	return TRUE;
}

// @@@ void INVENTORY::ClearLetter(symbol_type symbol, char letter)

void INVENTORY::ClearLetter(symbol_type symbol, char letter)
{
    if(isupper(letter))
	upper[SymbolToInt(symbol)] &= ~(1 << letter - 'A');
    else if(islower(letter))
	lower[SymbolToInt(symbol)] &= ~(1 << letter - 'a');
}


// @@@ char INVENTORY::ChooseLetter(symbol_type symbol, int letter = ERROR)

char INVENTORY::ChooseLetter(symbol_type symbol, int letter = ERROR)
{
    if(letter == ERROR || CheckLetter(symbol, letter))
	for(letter = FirstLetter;
	    letter != SYMBOL_BUNCH && CheckLetter(symbol, letter);
	    letter = NextLetter(letter))
	    /* Do Nothing */;

    SetLetter(symbol, letter);
    return letter;
}    

// @@@ void INVENTORY::create()

void INVENTORY::create()
{
    for(int i = 0; i < SymbolCount; i++)
	upper[i] = lower[i] = 0L;

    IsLit = FALSE;
    NextEvent = 0L;
    VisibleLook = SYMBOL_ERROR;
    TotalLook = SYMBOL_ERROR;
    IsTouched = TRUE;
    MyWeight = 0L;
    MySize = 0L;
}

// @@@ void INVENTORY::update(LEVEL* level, coor pos, MONSTER* monster = NULL)

void INVENTORY::update(LEVEL* level, coor pos, MONSTER* monster = NULL)
{
    long VisibleSize = 0L;
    long TotalSize = 0L;
    VisibleLook = SYMBOL_ERROR;
    TotalLook = SYMBOL_ERROR;
    NextEvent = 0L;
    IsLit = FALSE;
    MyWeight = 0L;
    MySize = 0L;

    reset();
    while(current() != NULL)
    {
	if(current()->event(level, pos, monster) == DONE)
	    RemoveObject();
	else
	{
	    long event = current()->NextEvent();

	    if(event != 0L && (NextEvent == 0L || event < NextEvent))
		NextEvent = event;

	    IsLit |= current()->light();
	    MyWeight += current()->weight();
	    MySize += current()->size();

	    if(current()->visible())
	    {
		if(current()->size() > VisibleSize)
		{
		    VisibleSize = current()->size();
		    VisibleLook = current()->look();
		}
	    }
	    if(current()->size() > TotalSize)
	    {
		TotalSize = current()->size();
		TotalLook = current()->look();
	    }
	    next();
	}
    }
    if(NextEvent != 0L)
	level->AddEvent(pos, NextEvent);

    IsTouched = FALSE;
}

// @@@ void INVENTORY::RemoveObject()

void INVENTORY::RemoveObject()
{
    if(IsKind(current()->symbol()))
	ClearLetter(current()->symbol(), current()->letter());
    
    list.remove();
    touch();
}

// @@@ void INVENTORY::AddObject(OBJECT* object)

void INVENTORY::AddObject(OBJECT* object)
{
    if(IsKind(object->symbol()))
	object->SetLetter(ChooseLetter(object->symbol(), object->letter()));

    list.append(object);
    touch();
}

// @@@ void INVENTORY::send(inventory_type type, PLAYER* player)
//
// Send an inventory description to the player.
// See also INVENTORY::CreatePage()

void INVENTORY::send(inventory_type type, PLAYER* player)
{
    // Sort inventory, and get concatenate equal objects
    CleanUp();

    PAGE FullPage;		// All information on the bunch

    reset();
    for(;;)
    {
	while(remain() > 0 && !IsKind(current()->symbol()))
	    next();

	if(remain() < 1)
	    break;
	
	char symbol = current()->symbol();
	PAGE page;
	
	FullPage.append(format("\n%c", symbol));
	FullPage.append(list.PackPage(symbol, &page, player));
	for(page.reset(); page.length() > 0; page.remove())
	    FullPage.append(page.current());
    }
    FullPage.append(STRING("", STRING_STATIC));

    player->SendBunch(type, &FullPage);
}

// @@@ void INVENTORY::add(OBJECT* object)

void INVENTORY::add(OBJECT* object)
{ AddObject(object); }

// @@@ void INVENTORY::add(OBJECT_LIST* list)

void INVENTORY::add(OBJECT_LIST* list)
{
    list->reset();
    while(list->current() != NULL)
    {
	AddObject(list->current());
	list->next();
    }
}

// @@@ STRING INVENTORY::add(OBJECT* object

STRING INVENTORY::add(OBJECT* object, print_flag flag, PLAYER* player = NULL)
{
    AddObject(object);
    return PrintObject(object, flag, player);
}

// @@@ void INVENTORY::add(OBJECT_LIST* list,

void INVENTORY::add(OBJECT_LIST* list, PAGE* page, PLAYER* player = NULL)
{
    OBJECT* object;

    // Assign Letters
    list->reset();
    while((object = list->current()) != NULL)
    {
	if(IsKind(object->symbol()))
	    object->SetLetter(ChooseLetter(object->symbol(),
					   object->letter()));
	
	list->next();
    }
    list->CreatePage(page, player);

    // Remove letters again
    list->reset();
    while((object = list->current()) != NULL)
    {
	if(IsKind(object->symbol()))
	    ClearLetter(object->symbol(), object->letter());
	
	list->next();
    }

    // Put 'em in for real.
    add(list);
}

// @@@ void INVENTORY::remove(OBJECT* object)

void INVENTORY::remove(OBJECT* object)
{
    reset();
    while(current() != NULL)
    {
	if(current() == object)
	{
	    RemoveObject();
	    return;
	}
	next();
    }
}

// @@@ void INVENTORY::remove(SPECIFY* specify, 
//
// See also: PACK::RemoveHalf()...

void INVENTORY::remove(SPECIFY* specify, OBJECT_LIST* list,
		       MONSTER* monster = NULL, long argument = 0L)
{
    if(argument > 0L)
    {
	long total = 0L;

	reset();
	while(current() != NULL)
	{
	    if(specify->check(current()) && current()->CanRemove(monster))
		total += current()->count();

	    next();
	}
	if(total > argument)
	{
	    reset();
	    while(current() != NULL && argument > 0)
	    {
		long count = current()->count();
		total -= count;

		if(specify->check(current()) &&
		   current()->CanRemove(monster) &&
		   rnd.chance((int) argument, (int) total + 1))
		{
		    if(argument >= count)
		    {
			list->append(current());
			argument -= count;
			RemoveObject();
		    }
		    else
		    {
			touch();
			list->append(current()->split(argument));
			return;
		    }
		}
		else
		    next();
	    }
	    return;
	}
    }
    reset();
    while(current() != NULL)
    {
	if(specify->check(current()) && (argument == -1L ||
					 current()->CanRemove(monster)))
					 
	{
	    list->append(current());
	    RemoveObject();
	}
	else
	    next();
    }
}

// @@@ bool INVENTORY::contain(SPECIFY* specify)

bool INVENTORY::contain(SPECIFY* specify)
{
    reset();

    while(current() != NULL)
    {
	if(specify->check(current()))
	    return TRUE;
	else
	    next();
    }
    return FALSE;
}

// @@@ void INVENTORY::free()
//
// Deallocate all objects in the inventory

void INVENTORY::free()
{
    reset();
    while(current() != NULL)
	list.erase();

    create();
}

// @@@ void INVENTORY::save(BINARY* binary)

void INVENTORY::save(BINARY* binary)
{
    list.save(binary);
}

// @@@ void INVENTORY::load(BINARY* binary)

void INVENTORY::load(BINARY* binary)
{
    OBJECT_LIST list;

    list.load(binary);
    add(&list);
}

// @@@ void INVENTORY::load(PARSE* parse, coor pos)
//
// Load objects from a text file

void INVENTORY::load(PARSE* parse, coor pos)
{
    OBJECT_LIST list;

    list.load(parse, pos);
    add(&list);
}

// @@@ INVENTORY::INVENTORY()

INVENTORY::INVENTORY()
{ create(); }

// @@@ INVENTORY::INVENTORY(PARSE* parse, coor pos)

INVENTORY::INVENTORY(PARSE* parse, coor pos)
{
    create();
    load(parse, pos);
}

// @@@ INVENTORY::~INVENTORY()

INVENTORY::~INVENTORY()
{ free(); }

// @@ emacs
//
// Local Variables:
// mode: c++
// outline-regexp: "// @+ "
// eval: (outline-minor-mode 1)
// End:
